"""Tests for OpenAPI customization (operation IDs, include_in_schema)."""

from __future__ import annotations

from collections.abc import AsyncGenerator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from tests.conftest import SessionFactory

from auen import CrudRouterBuilder, Operation


@pytest.fixture
def custom_opid_app(get_session: SessionFactory) -> FastAPI:
    application = FastAPI()
    application.include_router(
        CrudRouterBuilder.for_model(
            __import__("tests.conftest", fromlist=["Book"]).Book,
            get_session,
        )
        .with_operation_id_prefix("books_v2")
        .build()
    )
    return application


@pytest.fixture
async def custom_opid_client(
    custom_opid_app: FastAPI,
) -> AsyncGenerator[AsyncClient, None]:
    transport = ASGITransport(app=custom_opid_app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        yield c


async def test_custom_operation_ids(custom_opid_client: AsyncClient) -> None:
    resp = await custom_opid_client.get("/openapi.json")
    schema = resp.json()
    op_ids = []
    for path_item in schema["paths"].values():
        for method_info in path_item.values():
            if "operationId" in method_info:
                op_ids.append(method_info["operationId"])
    assert "books_v2_create" in op_ids
    assert "books_v2_nested_create" in op_ids
    assert "books_v2_list" in op_ids
    assert "books_v2_read" in op_ids
    assert "books_v2_update" in op_ids
    assert "books_v2_nested_update" in op_ids
    assert "books_v2_delete" in op_ids


@pytest.fixture
def hidden_ops_app(get_session: SessionFactory) -> FastAPI:
    from tests.conftest import Book

    application = FastAPI()
    application.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_include_in_schema({Operation.LIST, Operation.READ})
        .build()
    )
    return application


@pytest.fixture
async def hidden_ops_client(
    hidden_ops_app: FastAPI,
) -> AsyncGenerator[AsyncClient, None]:
    transport = ASGITransport(app=hidden_ops_app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        yield c


async def test_include_in_schema_hides_operations(
    hidden_ops_client: AsyncClient,
) -> None:
    """Operations not in include_in_schema should be absent from OpenAPI."""
    resp = await hidden_ops_client.get("/openapi.json")
    schema = resp.json()

    # The endpoints still work, but they're hidden from the schema
    methods_in_schema = set()
    for path_item in schema["paths"].values():
        methods_in_schema.update(path_item.keys())

    assert "get" in methods_in_schema  # LIST and READ are visible
    assert "post" not in methods_in_schema  # CREATE hidden
    assert "patch" not in methods_in_schema  # UPDATE hidden
    assert "delete" not in methods_in_schema  # DELETE hidden

    # But the endpoints still work
    resp = await hidden_ops_client.post("/books/", json={"title": "Test", "isbn": "123"})
    assert resp.status_code == 201



"""Direct async endpoint tests and endpoint utilities."""

from __future__ import annotations

from types import SimpleNamespace
from unittest.mock import AsyncMock

import pytest
from fastapi import APIRouter, HTTPException
from fastapi.routing import APIRoute
from pydantic import BaseModel, create_model
from sqlalchemy.exc import IntegrityError
from sqlmodel import Field, Relationship, SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession
from tests.conftest import Book, SessionFactory

from auen import (
    CrudRouterBuilder,
    FilterConfig,
    FilterFieldConfig,
    HooksConfig,
    Operation,
    PaginationConfig,
    derive_schemas,
)
from auen._async_endpoints import (
    _build_nested_create_schema,
    _build_nested_update_schema,
    _resolve_relation_fk_field,
)
from auen._endpoints import RouterContext, _build_filter_model
from auen.config import NestedRelationConfig, NestedWriteConfig
from auen.exceptions import (
    ConfigurationError,
    ConflictError,
    ForbiddenError,
    NotFoundError,
)
from auen.types import SelectQuery, User


class EndpointCovAuthor(SQLModel, table=True):
    __tablename__ = "endpoint_cov_author"
    id: int | None = Field(default=None, primary_key=True)
    name: str
    bio: str | None = None


class EndpointCovBook(SQLModel, table=True):
    __tablename__ = "endpoint_cov_book"
    id: int | None = Field(default=None, primary_key=True)
    title: str
    isbn: str
    author_id: int | None = Field(default=None, foreign_key="endpoint_cov_author.id")
    author: EndpointCovAuthor = Relationship()


class EndpointCovOther(SQLModel, table=True):
    __tablename__ = "endpoint_cov_other"
    id: int | None = Field(default=None, primary_key=True)
    label: str


class EndpointCovAllowAll:
    def can_create(self, user: User, obj_in: BaseModel) -> bool:
        return True

    def can_read(self, user: User, db_obj: SQLModel) -> bool:
        return True

    def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
        return True

    def can_delete(self, user: User, db_obj: SQLModel) -> bool:
        return True

    def filter_list_query(
        self, user: User, query: SelectQuery[SQLModel]
    ) -> SelectQuery[SQLModel]:
        return query


class EndpointCovDenyUpdate(EndpointCovAllowAll):
    def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
        return False


class EndpointCovDenyCreate(EndpointCovAllowAll):
    def can_create(self, user: User, obj_in: BaseModel) -> bool:
        return False


class EndpointCovRelationPolicyDenyCreate(EndpointCovAllowAll):
    def can_create(self, user: User, obj_in: BaseModel) -> bool:
        return False


class EndpointCovRelationPolicyDenyUpdate(EndpointCovAllowAll):
    def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
        return False


def _get_route(router: APIRouter, path: str, method: str) -> APIRoute:
    prefix = router.prefix or ""
    full_path = f"{prefix}{path}"
    for route in router.routes:
        if (
            isinstance(route, APIRoute)
            and route.path == full_path
            and method in route.methods
        ):
            return route
    msg = f"Route {method} {full_path} not found"
    raise AssertionError(msg)


async def test_direct_endpoints_roundtrip(get_session: SessionFactory) -> None:
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_pagination(PaginationConfig(include_total=True))
        .with_operations(
            {
                Operation.CREATE,
                Operation.READ,
                Operation.UPDATE,
                Operation.DELETE,
                Operation.LIST,
                Operation.BULK_CREATE,
                Operation.BULK_UPDATE,
                Operation.BULK_DELETE,
            }
        )
        .build()
    )

    schemas = derive_schemas(Book)
    create_ep = _get_route(router, "/", "POST").endpoint
    list_ep = _get_route(router, "/", "GET").endpoint
    read_ep = _get_route(router, "/{id}", "GET").endpoint
    update_ep = _get_route(router, "/{id}", "PATCH").endpoint
    delete_ep = _get_route(router, "/{id}", "DELETE").endpoint
    bulk_create_ep = _get_route(router, "/bulk", "POST").endpoint
    bulk_update_ep = _get_route(router, "/bulk", "PATCH").endpoint
    bulk_delete_ep = _get_route(router, "/bulk", "DELETE").endpoint

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="A", isbn="1"),
            session=session,
            user=None,
        )
        created_id = created.id
        await create_ep(
            obj_in=schemas.create(title="B", isbn="2"),
            session=session,
            user=None,
        )

        payload = await list_ep(
            session=session,
            user=None,
            offset=0,
            limit=10,
            sort=None,
            filter_params=None,
        )
        assert payload["total"] >= 2

        found = await read_ep(
            id=created_id,
            session=session,
            user=None,
        )
        assert found.id == created_id

        updated = await update_ep(
            id=created_id,
            obj_in=schemas.update(pages=123),
            session=session,
            user=None,
        )
        assert updated.pages == 123

        bulk_created = await bulk_create_ep(
            objects_in=[
                schemas.create(title="C", isbn="3"),
                schemas.create(title="D", isbn="4"),
            ],
            session=session,
            user=None,
        )
        bulk_ids = [obj.id for obj in bulk_created]
        # Send back ReadSchema-shaped objects
        bulk_update = await bulk_update_ep(
            items_in=[
                schemas.read(id=bulk_ids[0], title="C2", isbn="3", pages=None),
            ],
            session=session,
            user=None,
        )
        assert bulk_update[0].title == "C2"

        # Bulk delete uses identifier schema
        from pydantic import create_model as _cm

        BookIdentifier = _cm("BookIdentifier", __base__=BaseModel, id=(int, ...))
        await bulk_delete_ep(
            items_in=[BookIdentifier(id=bulk_ids[1])],
            session=session,
            user=None,
        )

        await delete_ep(id=created_id, session=session, user=None)


async def test_create_forbidden(get_session: SessionFactory) -> None:
    class DenyCreate:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return False

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return True

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyCreate())
        .with_operations({Operation.CREATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint

    async for session in get_session():
        with pytest.raises(ForbiddenError):
            await create_ep(
                obj_in=schemas.create(title="Nope", isbn="X"),
                session=session,
                user=None,
            )


async def test_create_after_hook(get_session: SessionFactory) -> None:
    schemas = derive_schemas(Book)
    called = []

    async def after_create(session: AsyncSession, obj: Book, user: User) -> None:
        called.append(obj.title)

    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(after_create=after_create))
        .with_operations({Operation.CREATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint

    async for session in get_session():
        await create_ep(
            obj_in=schemas.create(title="Hooked", isbn="H"),
            session=session,
            user=None,
        )

    assert called == ["Hooked"]


def test_build_filter_model_skips_missing_fields() -> None:
    cfg = FilterConfig(fields={"missing": FilterFieldConfig()})
    model = _build_filter_model(Book, cfg)
    assert model.model_fields == {}


async def test_list_without_total_returns_items(get_session: SessionFactory) -> None:
    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_pagination(PaginationConfig(include_total=False))
        .with_operations({Operation.CREATE, Operation.LIST})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    list_ep = _get_route(router, "/", "GET").endpoint
    async for session in get_session():
        await create_ep(
            obj_in=schemas.create(title="NoTotal", isbn="NT"),
            session=session,
            user=None,
        )
        payload = await list_ep(
            session=session,
            user=None,
            offset=0,
            limit=10,
            sort=None,
            filter_params=None,
        )
        assert isinstance(payload, list)


async def test_read_not_found_and_forbidden(get_session: SessionFactory) -> None:
    class DenyRead:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return True

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return False

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return True

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyRead())
        .with_operations({Operation.CREATE, Operation.READ})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    read_ep = _get_route(router, "/{id}", "GET").endpoint
    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="Denied", isbn="DR"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await read_ep(id=created.id, session=session, user=None)
        with pytest.raises(NotFoundError):
            await read_ep(id=9999, session=session, user=None)


async def test_update_branches(get_session: SessionFactory) -> None:
    class DenyUpdate:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return True

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return False

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    called = []

    async def before_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before")

    async def after_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after")

    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(before_update=before_update, after_update=after_update))
        .with_operations({Operation.CREATE, Operation.UPDATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    update_ep = _get_route(router, "/{id}", "PATCH").endpoint
    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="Up", isbn="U"),
            session=session,
            user=None,
        )
        result = await update_ep(
            id=created.id,
            obj_in=schemas.update(title="Up2"),
            session=session,
            user=None,
        )
        assert result.title == "Up2"
        assert called == ["before", "after"]

    router_forbidden = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyUpdate())
        .with_operations({Operation.CREATE, Operation.UPDATE})
        .build()
    )
    create_forbidden = _get_route(router_forbidden, "/", "POST").endpoint
    update_forbidden = _get_route(router_forbidden, "/{id}", "PATCH").endpoint
    async for session in get_session():
        created = await create_forbidden(
            obj_in=schemas.create(title="DeniedUp", isbn="DU"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await update_forbidden(
                id=created.id,
                obj_in=schemas.update(title="x"),
                session=session,
                user=None,
            )
        with pytest.raises(NotFoundError):
            await update_forbidden(
                id=9999,
                obj_in=schemas.update(title="x"),
                session=session,
                user=None,
            )


async def test_delete_branches(get_session: SessionFactory) -> None:
    class DenyDelete:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return True

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return True

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return False

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    called = []

    async def before_delete(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before")

    async def after_delete(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after")

    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(before_delete=before_delete, after_delete=after_delete))
        .with_operations({Operation.CREATE, Operation.DELETE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    delete_ep = _get_route(router, "/{id}", "DELETE").endpoint
    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="Del", isbn="D"),
            session=session,
            user=None,
        )
        await delete_ep(id=created.id, session=session, user=None)
        assert called == ["before", "after"]

    router_forbidden = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyDelete())
        .with_operations({Operation.CREATE, Operation.DELETE})
        .build()
    )
    create_forbidden = _get_route(router_forbidden, "/", "POST").endpoint
    delete_forbidden = _get_route(router_forbidden, "/{id}", "DELETE").endpoint
    async for session in get_session():
        created = await create_forbidden(
            obj_in=schemas.create(title="DeniedDel", isbn="DD"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await delete_forbidden(id=created.id, session=session, user=None)
        with pytest.raises(NotFoundError):
            await delete_forbidden(id=9999, session=session, user=None)


async def test_bulk_create_forbidden(get_session: SessionFactory) -> None:
    class DenyCreate:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return False

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return True

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyCreate())
        .with_operations({Operation.BULK_CREATE})
        .build()
    )
    bulk_create_ep = _get_route(router, "/bulk", "POST").endpoint
    async for session in get_session():
        with pytest.raises(ForbiddenError):
            await bulk_create_ep(
                objects_in=[derive_schemas(Book).create(title="T", isbn="I")],
                session=session,
                user=None,
            )


async def test_bulk_update_forbidden_and_not_found(
    get_session: SessionFactory,
) -> None:
    class DenyUpdate:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return True

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return False

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_operations({Operation.CREATE, Operation.BULK_UPDATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    bulk_update_ep = _get_route(router, "/bulk", "PATCH").endpoint

    BulkUpdateItem = create_model(
        "BulkUpdateForbidden",
        __base__=BaseModel,
        id=(int, ...),
        title=(str | None, None),
    )

    async for session in get_session():
        await create_ep(
            obj_in=schemas.create(title="Bulk", isbn="B"),
            session=session,
            user=None,
        )
        with pytest.raises(NotFoundError):
            await bulk_update_ep(
                items_in=[BulkUpdateItem(id=9999, title="X")],
                session=session,
                user=None,
            )

    router_forbidden = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyUpdate())
        .with_operations({Operation.CREATE, Operation.BULK_UPDATE})
        .build()
    )
    create_ep2 = _get_route(router_forbidden, "/", "POST").endpoint
    bulk_update_forbidden = _get_route(router_forbidden, "/bulk", "PATCH").endpoint
    async for session in get_session():
        created = await create_ep2(
            obj_in=schemas.create(title="Bulk2", isbn="B2"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await bulk_update_forbidden(
                items_in=[BulkUpdateItem(id=created.id, title="X")],
                session=session,
                user=None,
            )


async def test_bulk_delete_branches(get_session: SessionFactory) -> None:
    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_operations({Operation.CREATE, Operation.BULK_DELETE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    bulk_delete_ep = _get_route(router, "/bulk", "DELETE").endpoint

    from pydantic import create_model as _cm

    BookIdentifier = _cm("BookIdentifier", __base__=BaseModel, id=(int, ...))

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="SB", isbn="SB1"),
            session=session,
            user=None,
        )
        await bulk_delete_ep(
            items_in=[BookIdentifier(id=created.id)], session=session, user=None
        )
        with pytest.raises(NotFoundError):
            await bulk_delete_ep(
                items_in=[BookIdentifier(id=9999)], session=session, user=None
            )


async def test_bulk_delete_forbidden(get_session: SessionFactory) -> None:
    class DenyDelete:
        def can_create(self, user: User, obj_in: BaseModel) -> bool:
            return True

        def can_read(self, user: User, db_obj: SQLModel) -> bool:
            return True

        def can_update(self, user: User, db_obj: SQLModel, obj_in: BaseModel) -> bool:
            return True

        def can_delete(self, user: User, db_obj: SQLModel) -> bool:
            return False

        def filter_list_query(
            self, user: User, query: SelectQuery[SQLModel]
        ) -> SelectQuery[SQLModel]:
            return query

    schemas = derive_schemas(Book)
    router = (
        CrudRouterBuilder.for_model(Book, get_session)
        .with_policy(DenyDelete())
        .with_operations({Operation.CREATE, Operation.BULK_DELETE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    bulk_delete_ep = _get_route(router, "/bulk", "DELETE").endpoint
    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="FD", isbn="FD"),
            session=session,
            user=None,
        )
        from pydantic import create_model as _cm2

        BookId = _cm2("BookIdentifier", __base__=BaseModel, id=(int, ...))
        with pytest.raises(ForbiddenError):
            await bulk_delete_ep(
                items_in=[BookId(id=created.id)], session=session, user=None
            )


def test_resolve_relation_fk_field_unknown_relation() -> None:
    with pytest.raises(ConfigurationError, match="Unknown relationship"):
        _resolve_relation_fk_field(
            EndpointCovBook,
            "unknown",
            EndpointCovAuthor,
        )


def test_resolve_relation_fk_field_not_many_to_one(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from sqlalchemy.orm import RelationshipDirection

    fake_rel = SimpleNamespace(
        direction=RelationshipDirection.ONETOMANY,
        mapper=SimpleNamespace(class_=EndpointCovAuthor),
        local_columns=(SimpleNamespace(name="author_id"),),
    )
    fake_mapper = SimpleNamespace(relationships={"author": fake_rel})
    monkeypatch.setattr("auen._async_endpoints.sa_inspect", lambda _model: fake_mapper)

    with pytest.raises(ConfigurationError, match="MANYTOONE"):
        _resolve_relation_fk_field(
            EndpointCovBook,
            "author",
            EndpointCovAuthor,
        )


def test_resolve_relation_fk_field_model_mismatch(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from sqlalchemy.orm import RelationshipDirection

    fake_rel = SimpleNamespace(
        direction=RelationshipDirection.MANYTOONE,
        mapper=SimpleNamespace(class_=EndpointCovAuthor),
        local_columns=(SimpleNamespace(name="author_id"),),
    )
    fake_mapper = SimpleNamespace(relationships={"author": fake_rel})
    monkeypatch.setattr("auen._async_endpoints.sa_inspect", lambda _model: fake_mapper)

    with pytest.raises(ConfigurationError, match="model mismatch"):
        _resolve_relation_fk_field(
            EndpointCovBook,
            "author",
            EndpointCovBook,
        )


def test_resolve_relation_fk_field_multiple_local_columns(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from sqlalchemy.orm import RelationshipDirection

    fake_rel = SimpleNamespace(
        direction=RelationshipDirection.MANYTOONE,
        mapper=SimpleNamespace(class_=EndpointCovAuthor),
        local_columns=(
            SimpleNamespace(name="author_id"),
            SimpleNamespace(name="author_id2"),
        ),
    )
    fake_mapper = SimpleNamespace(relationships={"author": fake_rel})
    monkeypatch.setattr("auen._async_endpoints.sa_inspect", lambda _model: fake_mapper)

    with pytest.raises(ConfigurationError, match="exactly one local FK"):
        _resolve_relation_fk_field(
            EndpointCovBook,
            "author",
            EndpointCovAuthor,
        )


def test_resolve_relation_fk_field_success() -> None:
    assert (
        _resolve_relation_fk_field(EndpointCovBook, "author", EndpointCovAuthor)
        == "author_id"
    )


def test_build_nested_update_schema_fallback_derived() -> None:
    base = derive_schemas(EndpointCovBook)
    ctx = RouterContext(
        router=APIRouter(),
        model=EndpointCovBook,
        model_name=EndpointCovBook.__name__,
        schemas=base.__class__(
            create=base.create,
            read=base.read,
            update=base.update,
            nested_update=None,
        ),
        policy=EndpointCovAllowAll(),
        session_dep=lambda: None,
        auth=None,
        pk_name="id",
        pk_type=int,
        pagination=PaginationConfig(),
        filters=None,
        repository_factory=lambda _model: None,
        hooks=HooksConfig(),
        nested_writes=None,
        nested_policy_registry={},
        include_in_schema={Operation.NESTED_UPDATE},
        operation_id_prefix="book",
    )
    nested_schema = _build_nested_update_schema(ctx)
    assert nested_schema is not None
    assert "title" in nested_schema.model_fields


def test_build_nested_create_schema_fallback_derived() -> None:
    base = derive_schemas(EndpointCovBook)
    ctx = RouterContext(
        router=APIRouter(),
        model=EndpointCovBook,
        model_name=EndpointCovBook.__name__,
        schemas=base.__class__(
            create=base.create,
            read=base.read,
            update=base.update,
            nested_create=None,
            nested_update=base.nested_update,
        ),
        policy=EndpointCovAllowAll(),
        session_dep=lambda: None,
        auth=None,
        pk_name="id",
        pk_type=int,
        pagination=PaginationConfig(),
        filters=None,
        repository_factory=lambda _model: None,
        hooks=HooksConfig(),
        nested_writes=None,
        nested_policy_registry={},
        include_in_schema={Operation.NESTED_CREATE},
        operation_id_prefix="book",
    )
    nested_schema = _build_nested_create_schema(ctx)
    assert nested_schema is not None
    assert "title" in nested_schema.model_fields


async def test_nested_create_forbidden_validation_and_success_branches(
    get_session: SessionFactory,
) -> None:
    called: list[str] = []

    async def before_create(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before")

    async def after_create(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after")

    nested_cfg = NestedWriteConfig(
        fields={"author": NestedRelationConfig(model=EndpointCovAuthor)}
    )
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_hooks(HooksConfig(before_create=before_create, after_create=after_create))
        .with_nested_writes(nested_cfg)
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        created = await nested_create_ep(
            obj_in=nested_obj_type(title="B1", isbn="1", author={"name": "A1"}),
            session=session,
            user=None,
        )
        assert created.title == "B1"
        assert created.author_id is not None

        with pytest.raises(HTTPException, match="must be an object"):
            await nested_create_ep(
                obj_in=nested_obj_type(title="B2", isbn="2", author=None),
                session=session,
                user=None,
            )

        with pytest.raises(HTTPException, match="Field required"):
            await nested_create_ep(
                obj_in=nested_obj_type(title="B3", isbn="3", author={}),
                session=session,
                user=None,
            )

        author = EndpointCovAuthor(name="A-existing", bio="old")
        session.add(author)
        await session.flush()
        author_id = author.id

        updated = await nested_create_ep(
            obj_in=nested_obj_type(
                title="B4",
                isbn="4",
                author={"id": author_id, "name": "A-updated", "bio": "new"},
            ),
            session=session,
            user=None,
        )
        assert updated.author_id == author_id

        refreshed = await session.get(EndpointCovAuthor, author_id)
        assert refreshed is not None
        assert refreshed.name == "A-updated"
        assert called.count("before") == 4
        assert called.count("after") == 2

    forbidden_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_policy(EndpointCovDenyCreate())
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(forbidden_router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        with pytest.raises(ForbiddenError):
            await nested_create_ep(
                obj_in=nested_obj_type(title="Denied", isbn="D"),
                session=session,
                user=None,
            )


async def test_nested_create_relation_policy_denies_create_and_update(
    get_session: SessionFactory,
) -> None:
    deny_create_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={
                    "author": NestedRelationConfig(
                        model=EndpointCovAuthor,
                        policy=EndpointCovRelationPolicyDenyCreate(),
                    )
                }
            )
        )
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(deny_create_router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        with pytest.raises(ForbiddenError):
            await nested_create_ep(
                obj_in=nested_obj_type(title="C", isbn="1", author={"name": "blocked"}),
                session=session,
                user=None,
            )

    deny_update_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={
                    "author": NestedRelationConfig(
                        model=EndpointCovAuthor,
                        policy=EndpointCovRelationPolicyDenyUpdate(),
                    )
                }
            )
        )
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(deny_update_router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        author = EndpointCovAuthor(name="A")
        session.add(author)
        await session.flush()
        author_id = author.id

        with pytest.raises(ForbiddenError):
            await nested_create_ep(
                obj_in=nested_obj_type(
                    title="C2",
                    isbn="2",
                    author={"id": author_id, "name": "blocked"},
                ),
                session=session,
                user=None,
            )


async def test_nested_create_invalid_nested_writes_configs(
    get_session: SessionFactory,
) -> None:
    unknown_relation_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={"missing": NestedRelationConfig(model=EndpointCovAuthor)}
            )
        )
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(unknown_relation_router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        with pytest.raises(ConfigurationError, match="Unknown relationship 'missing'"):
            await nested_create_ep(
                obj_in=nested_obj_type(title="X", isbn="1"),
                session=session,
                user=None,
            )

    model_mismatch_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={"author": NestedRelationConfig(model=EndpointCovOther)}
            )
        )
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(model_mismatch_router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    async for session in get_session():
        with pytest.raises(ConfigurationError, match="model mismatch"):
            await nested_create_ep(
                obj_in=nested_obj_type(title="Y", isbn="2"),
                session=session,
                user=None,
            )


async def test_nested_create_conflict_and_no_after_hook_branches(
    get_session: SessionFactory,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_operations({Operation.NESTED_CREATE})
        .build()
    )
    nested_create_ep = _get_route(router, "/nested", "POST").endpoint
    nested_obj_type = nested_create_ep.__annotations__["obj_in"]

    fake_obj = EndpointCovBook(id=1, title="B", isbn="1")
    monkeypatch.setattr(
        "auen._async_endpoints.apply_nested_creates",
        AsyncMock(return_value=fake_obj),
    )

    ok_session = SimpleNamespace(
        commit=AsyncMock(),
        refresh=AsyncMock(),
        rollback=AsyncMock(),
    )
    result = await nested_create_ep(
        obj_in=nested_obj_type(title="B", isbn="1"),
        session=ok_session,
        user=None,
    )
    assert result is fake_obj

    failing_session = SimpleNamespace(
        commit=AsyncMock(
            side_effect=IntegrityError("stmt", {"x": "y"}, Exception("boom"))
        ),
        refresh=AsyncMock(),
        rollback=AsyncMock(),
    )
    with pytest.raises(ConflictError):
        await nested_create_ep(
            obj_in=nested_obj_type(title="B2", isbn="2"),
            session=failing_session,
            user=None,
        )
    assert failing_session.rollback.await_count == 1
    assert failing_session.refresh.await_count == 0


async def test_nested_update_not_found_forbidden_and_plain_update_path(
    get_session: SessionFactory,
) -> None:
    schemas = derive_schemas(EndpointCovBook)
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    nested_update_ep = _get_route(router, "/{id}/nested", "PATCH").endpoint

    async for session in get_session():
        with pytest.raises(NotFoundError):
            await nested_update_ep(
                id=9999,
                obj_in=schemas.update(title="x"),
                session=session,
                user=None,
            )

        created = await create_ep(
            obj_in=schemas.create(title="Plain", isbn="P"),
            session=session,
            user=None,
        )
        result = await nested_update_ep(
            id=created.id,
            obj_in=schemas.update(title="Plain2"),
            session=session,
            user=None,
        )
        assert result.title == "Plain2"

    forbidden_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_policy(EndpointCovDenyUpdate())
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(forbidden_router, "/", "POST").endpoint
    nested_update_ep = _get_route(forbidden_router, "/{id}/nested", "PATCH").endpoint

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="Denied", isbn="D"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await nested_update_ep(
                id=created.id,
                obj_in=schemas.update(title="Nope"),
                session=session,
                user=None,
            )


async def test_nested_update_validation_and_success_branches(
    get_session: SessionFactory,
) -> None:
    called: list[str] = []

    async def before_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before")

    async def after_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after")

    nested_cfg = NestedWriteConfig(
        fields={"author": NestedRelationConfig(model=EndpointCovAuthor)}
    )
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_hooks(HooksConfig(before_update=before_update, after_update=after_update))
        .with_nested_writes(nested_cfg)
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    schemas = derive_schemas(EndpointCovBook)
    create_ep = _get_route(router, "/", "POST").endpoint
    nested_update_ep = _get_route(router, "/{id}/nested", "PATCH").endpoint
    nested_obj_type = nested_update_ep.__annotations__["obj_in"]

    async for session in get_session():
        author = EndpointCovAuthor(name="A", bio="old")
        author2 = EndpointCovAuthor(name="B", bio="old")
        session.add(author)
        session.add(author2)
        await session.flush()
        author_id = author.id
        author2_id = author2.id

        created = await create_ep(
            obj_in=schemas.create(title="B", isbn="1", author_id=author_id),
            session=session,
            user=None,
        )

        plain_updated = await nested_update_ep(
            id=created.id,
            obj_in=nested_obj_type(title="B0"),
            session=session,
            user=None,
        )
        assert plain_updated.title == "B0"

        with pytest.raises(HTTPException, match="must be an object"):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(author=None),
                session=session,
                user=None,
            )

        with pytest.raises(HTTPException, match="without nested PK"):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(author_id=author_id, author={"name": "n"}),
                session=session,
                user=None,
            )

        with pytest.raises(HTTPException, match="Conflict between"):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(
                    author_id=author_id,
                    author={"id": author2_id, "name": "n"},
                ),
                session=session,
                user=None,
            )

        with pytest.raises(NotFoundError):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(author={"id": 9999, "name": "missing"}),
                session=session,
                user=None,
            )

        updated = await nested_update_ep(
            id=created.id,
            obj_in=nested_obj_type(
                title="B2",
                author={"id": author_id, "name": "A2", "bio": "new"},
            ),
            session=session,
            user=None,
        )
        assert updated.title == "B2"
        assert updated.author_id == author_id

        same_fk_updated = await nested_update_ep(
            id=created.id,
            obj_in=nested_obj_type(
                author_id=author_id,
                author={"id": author_id, "bio": "same"},
            ),
            session=session,
            user=None,
        )
        assert same_fk_updated.author_id == author_id

        updated_book = await nested_update_ep(
            id=created.id,
            obj_in=nested_obj_type(author={"name": "Created", "bio": "created"}),
            session=session,
            user=None,
        )
        assert updated_book.author_id is not None
        assert updated_book.author_id not in {author_id, author2_id}
        assert called.count("before") == 8
        assert called.count("after") == 4
        assert called[-2:] == ["before", "after"]


async def test_nested_update_relation_policy_denies_create_and_update(
    get_session: SessionFactory,
) -> None:
    schemas = derive_schemas(EndpointCovBook)

    deny_create_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={
                    "author": NestedRelationConfig(
                        model=EndpointCovAuthor,
                        policy=EndpointCovRelationPolicyDenyCreate(),
                    )
                }
            )
        )
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(deny_create_router, "/", "POST").endpoint
    nested_update_ep = _get_route(deny_create_router, "/{id}/nested", "PATCH").endpoint
    nested_obj_type = nested_update_ep.__annotations__["obj_in"]

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="C", isbn="1"),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(author={"name": "blocked"}),
                session=session,
                user=None,
            )

    deny_update_router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={
                    "author": NestedRelationConfig(
                        model=EndpointCovAuthor,
                        policy=EndpointCovRelationPolicyDenyUpdate(),
                    )
                }
            )
        )
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(deny_update_router, "/", "POST").endpoint
    nested_update_ep = _get_route(deny_update_router, "/{id}/nested", "PATCH").endpoint
    nested_obj_type = nested_update_ep.__annotations__["obj_in"]

    async for session in get_session():
        author = EndpointCovAuthor(name="A")
        session.add(author)
        await session.flush()
        author_id = author.id

        created = await create_ep(
            obj_in=schemas.create(title="C2", isbn="2", author_id=author_id),
            session=session,
            user=None,
        )
        with pytest.raises(ForbiddenError):
            await nested_update_ep(
                id=created.id,
                obj_in=nested_obj_type(author={"id": author_id, "name": "blocked"}),
                session=session,
                user=None,
            )


async def test_nested_update_invalid_nested_writes_unknown_relation(
    get_session: SessionFactory,
) -> None:
    schemas = derive_schemas(EndpointCovBook)
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={"missing": NestedRelationConfig(model=EndpointCovAuthor)}
            )
        )
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    nested_update_ep = _get_route(router, "/{id}/nested", "PATCH").endpoint

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="C3", isbn="3"),
            session=session,
            user=None,
        )
        with pytest.raises(ConfigurationError, match="Unknown relationship 'missing'"):
            await nested_update_ep(
                id=created.id,
                obj_in=schemas.update(title="x"),
                session=session,
                user=None,
            )


async def test_nested_update_invalid_nested_writes_model_mismatch(
    get_session: SessionFactory,
) -> None:
    schemas = derive_schemas(EndpointCovBook)
    router = (
        CrudRouterBuilder.for_model(EndpointCovBook, get_session)
        .with_nested_writes(
            NestedWriteConfig(
                fields={"author": NestedRelationConfig(model=EndpointCovOther)}
            )
        )
        .with_operations({Operation.CREATE, Operation.NESTED_UPDATE})
        .build()
    )
    create_ep = _get_route(router, "/", "POST").endpoint
    nested_update_ep = _get_route(router, "/{id}/nested", "PATCH").endpoint

    async for session in get_session():
        created = await create_ep(
            obj_in=schemas.create(title="C4", isbn="4"),
            session=session,
            user=None,
        )
        with pytest.raises(ConfigurationError, match="model mismatch"):
            await nested_update_ep(
                id=created.id,
                obj_in=schemas.update(title="x"),
                session=session,
                user=None,
            )



"""Bulk operation endpoint tests."""

from __future__ import annotations

from collections.abc import AsyncGenerator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from tests.conftest import Book, SessionFactory

from auen import CrudRouterBuilder, Operation


@pytest.fixture
def bulk_app(get_session: SessionFactory) -> FastAPI:
    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_operations(
            {
                Operation.CREATE,
                Operation.READ,
                Operation.BULK_CREATE,
                Operation.BULK_UPDATE,
                Operation.BULK_DELETE,
            }
        )
        .build()
    )
    return app


@pytest.fixture
async def bulk_client(bulk_app: FastAPI) -> AsyncGenerator[AsyncClient, None]:
    transport = ASGITransport(app=bulk_app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        yield c


async def test_bulk_create(bulk_client: AsyncClient) -> None:
    resp = await bulk_client.post(
        "/books/bulk",
        json=[
            {"title": "Book A", "isbn": "111"},
            {"title": "Book B", "isbn": "222"},
        ],
    )
    assert resp.status_code == 201
    data = resp.json()
    assert len(data) == 2
    assert data[0]["title"] == "Book A"
    assert data[1]["title"] == "Book B"
    assert data[0]["id"] is not None


async def test_bulk_create_empty(bulk_client: AsyncClient) -> None:
    resp = await bulk_client.post("/books/bulk", json=[])
    assert resp.status_code == 201
    assert resp.json() == []


async def test_bulk_update(bulk_client: AsyncClient) -> None:
    # Create books first
    resp = await bulk_client.post(
        "/books/bulk",
        json=[
            {"title": "Old A", "isbn": "111"},
            {"title": "Old B", "isbn": "222"},
        ],
    )
    books = resp.json()

    # Send back full ReadSchema objects with changed fields
    books[0]["title"] = "New A"
    books[1]["title"] = "New B"
    resp = await bulk_client.patch(
        "/books/bulk",
        json=books,
    )
    assert resp.status_code == 200
    data = resp.json()
    assert data[0]["title"] == "New A"
    assert data[1]["title"] == "New B"


async def test_bulk_update_not_found(bulk_client: AsyncClient) -> None:
    resp = await bulk_client.patch(
        "/books/bulk",
        json=[{"id": 9999, "title": "Nope", "isbn": "X", "pages": None}],
    )
    assert resp.status_code == 404


async def test_bulk_delete(bulk_client: AsyncClient) -> None:
    resp = await bulk_client.post(
        "/books/bulk",
        json=[
            {"title": "Del A", "isbn": "333"},
            {"title": "Del B", "isbn": "444"},
        ],
    )
    books = resp.json()

    resp = await bulk_client.request(
        "DELETE",
        "/books/bulk",
        json=[{"id": books[0]["id"]}, {"id": books[1]["id"]}],
    )
    assert resp.status_code == 200

    # Verify deleted
    for book in books:
        resp = await bulk_client.get(f"/books/{book['id']}")
        assert resp.status_code == 404


async def test_bulk_delete_not_found(bulk_client: AsyncClient) -> None:
    resp = await bulk_client.request(
        "DELETE",
        "/books/bulk",
        json=[{"id": 9999}],
    )
    assert resp.status_code == 404



"""Quickstart example as a test - proves the 10-line setup works."""

from __future__ import annotations

from collections.abc import AsyncGenerator

from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.pool import StaticPool
from sqlmodel import Field, SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

from auen import CrudRouterBuilder, derive_schemas


async def test_quickstart() -> None:
    """The minimal quickstart: model + engine + fluent builder."""

    class Item(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        title: str
        done: bool = False

    engine = create_async_engine(
        "sqlite+aiosqlite://",
        echo=False,
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
    )
    async with engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)

    async def get_session() -> AsyncGenerator[AsyncSession, None]:
        async with AsyncSession(engine) as session:
            yield session

    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Item, get_session)
        .with_schemas(derive_schemas(Item))
        .build()
    )

    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        # Create
        resp = await client.post("/items/", json={"title": "Buy milk", "done": False})
        assert resp.status_code == 201
        item = resp.json()
        assert item["title"] == "Buy milk"
        item_id = item["id"]

        # Read
        resp = await client.get(f"/items/{item_id}")
        assert resp.status_code == 200
        assert resp.json()["title"] == "Buy milk"

        # List
        resp = await client.get("/items/")
        assert resp.status_code == 200
        assert len(resp.json()) == 1

        # Update
        resp = await client.patch(f"/items/{item_id}", json={"done": True})
        assert resp.status_code == 200
        assert resp.json()["done"] is True
        assert resp.json()["title"] == "Buy milk"  # unchanged

        # Delete
        resp = await client.delete(f"/items/{item_id}")
        assert resp.status_code == 200

        # Verify gone
        resp = await client.get(f"/items/{item_id}")
        assert resp.status_code == 404

    await engine.dispose()



"""Test that the README quickstart example works end-to-end."""

from __future__ import annotations

from collections.abc import AsyncGenerator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.pool import StaticPool
from sqlmodel import Field, SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

from auen import CrudRouterBuilder


class Author(SQLModel, table=True):
    __tablename__ = "readme_author"
    id: int | None = Field(default=None, primary_key=True)
    name: str
    bio: str | None = None


class Book(SQLModel, table=True):
    __tablename__ = "readme_book"
    id: int | None = Field(default=None, primary_key=True)
    title: str
    isbn: str
    author_id: int | None = Field(default=None, foreign_key="readme_author.id")


@pytest.fixture
async def readme_app() -> AsyncGenerator[FastAPI, None]:
    engine = create_async_engine(
        "sqlite+aiosqlite://",
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
    )
    async with engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)

    async def get_session() -> AsyncGenerator[AsyncSession, None]:
        async with AsyncSession(engine) as session:
            yield session

    app = FastAPI()
    for router in CrudRouterBuilder.for_model([Author, Book], get_session).build_all():
        app.include_router(router)

    yield app
    await engine.dispose()


async def test_readme_quickstart_roundtrip(readme_app: FastAPI) -> None:
    transport = ASGITransport(app=readme_app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        # Create an author
        resp = await c.post(
            "/authors/",
            json={"name": "Robert C. Martin", "bio": "Author of Clean Code"},
        )
        assert resp.status_code == 201
        author = resp.json()
        author_id = author["id"]
        assert author["name"] == "Robert C. Martin"

        # Create a book linked to the author
        resp = await c.post(
            "/books/",
            json={
                "title": "Clean Code",
                "isbn": "978-0132350884",
                "author_id": author_id,
            },
        )
        assert resp.status_code == 201
        book = resp.json()
        assert book["title"] == "Clean Code"
        assert book["author_id"] == author_id

        # List authors
        resp = await c.get("/authors/")
        assert resp.status_code == 200
        assert len(resp.json()) == 1

        # List books
        resp = await c.get("/books/")
        assert resp.status_code == 200
        assert len(resp.json()) == 1

        # Read single book
        resp = await c.get(f"/books/{book['id']}")
        assert resp.status_code == 200
        assert resp.json()["isbn"] == "978-0132350884"

        # Update book
        resp = await c.patch(
            f"/books/{book['id']}",
            json={"title": "Clean Code: A Handbook of Agile Software Craftsmanship"},
        )
        assert resp.status_code == 200
        assert resp.json()["title"].startswith("Clean Code: A Handbook")

        # Delete book
        resp = await c.delete(f"/books/{book['id']}")
        assert resp.status_code == 200

        # Verify deleted
        resp = await c.get(f"/books/{book['id']}")
        assert resp.status_code == 404



"""Shared endpoint utilities and context for CRUD operations."""

from collections.abc import Callable
from typing import Any, cast

from fastapi import APIRouter, Depends, Security
from fastapi.params import Depends as DependsParam
from fastapi.params import Security as SecurityParam
from pydantic import BaseModel, ConfigDict, create_model
from pydantic.dataclasses import dataclass
from sqlmodel import SQLModel

from auen.config import (
    AuthConfig,
    FilterConfig,
    FilterOp,
    HooksConfig,
    NestedWriteConfig,
    Operation,
    PaginationConfig,
    SchemaConfig,
)
from auen.policy import CrudPolicy
from auen.repository import AsyncCrudRepository
from auen.types import ModelFieldDefinition, PKType, SessionDep


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class RouterContext:
    """Shared state for endpoint builders."""

    router: APIRouter
    model: type[SQLModel]
    model_name: str
    schemas: SchemaConfig
    policy: CrudPolicy
    session_dep: Callable[..., SessionDep]
    auth: AuthConfig | None
    pk_name: str
    pk_type: PKType
    pagination: PaginationConfig
    filters: FilterConfig | None
    repository_factory: Callable[[type[SQLModel]], AsyncCrudRepository]
    hooks: HooksConfig
    nested_writes: NestedWriteConfig | None
    nested_policy_registry: dict[type[SQLModel], CrudPolicy]
    include_in_schema: set[Operation]
    operation_id_prefix: str


def _make_user_dep(
    auth: AuthConfig | None, op: Operation
) -> DependsParam | SecurityParam:
    """Build a unified user dependency for an operation."""
    if auth is None or op in auth.allow_anonymous:

        def _anon() -> None:
            return None

        return Depends(_anon)
    if auth.mode == "security":
        scopes = auth.scopes_by_operation.get(op, [])
        return Security(auth.dependency, scopes=scopes)
    return Depends(auth.dependency)


def _build_filter_model(
    model: type[SQLModel], filter_config: FilterConfig
) -> type[BaseModel]:
    fields: dict[str, ModelFieldDefinition] = {}
    for name, field_cfg in filter_config.fields.items():
        field_info = model.model_fields.get(name)
        if field_info is None:
            continue
        annotation = field_info.annotation or type(None)
        for op in sorted(field_cfg.ops, key=lambda o: o.value):
            param_name = name if op == FilterOp.EQ else f"{name}__{op.value}"
            if op in {FilterOp.IN, FilterOp.CONTAINS, FilterOp.ICONTAINS}:
                param_type = str | None
            else:
                param_type = annotation | None
            fields[param_name] = (cast(type, param_type), None)
    return create_model(
        f"{model.__name__}FilterParams",
        __base__=BaseModel,
        **cast(dict[str, Any], fields),
    )


def _build_page_model(read_schema: type[BaseModel]) -> type[BaseModel]:
    items_type = list.__class_getitem__(read_schema)
    return create_model(
        f"{read_schema.__name__}Page",
        __base__=BaseModel,
        items=(items_type, ...),
        total=(int, ...),
        limit=(int, ...),
        offset=(int, ...),
    )


def _build_identifier_schema(
    model_name: str,
    pk_name: str,
    pk_type: type,
) -> type[BaseModel]:
    """Build a minimal schema containing only the PK field."""
    _create_model = cast(Callable[..., type[BaseModel]], create_model)
    return _create_model(
        f"{model_name}Identifier",
        __base__=BaseModel,
        **{pk_name: (pk_type, ...)},
    )



"""Schema derivation helpers."""

from __future__ import annotations

from collections.abc import Callable
from functools import lru_cache
from types import GenericAlias
from typing import cast

from pydantic import BaseModel, ConfigDict, create_model
from pydantic.dataclasses import dataclass
from sqlmodel import SQLModel

from auen._query import resolve_pk_field
from auen._relations import RelationInfo, inspect_relations
from auen.config import NestedWriteConfig, SchemaConfig
from auen.types import ModelFieldDefinition


class _StrictInputModel(BaseModel):
    """Base model for request schemas that rejects unknown fields."""

    model_config = ConfigDict(extra="forbid")


@dataclass(frozen=True)
class SchemaContainer:
    """Expanded schema bundle including nested create/update schemas."""

    create: type[BaseModel]
    read: type[BaseModel]
    update: type[BaseModel]
    nested_create: type[BaseModel]
    nested_update: type[BaseModel]

    def to_config(self) -> SchemaConfig:
        return SchemaConfig(
            create=self.create,
            read=self.read,
            update=self.update,
            nested_create=self.nested_create,
            nested_update=self.nested_update,
        )


def _get_pk_field_names(model: type[SQLModel]) -> set[str]:
    """Get primary key field names from a SQLModel."""
    from sqlalchemy import inspect as sa_inspect

    try:
        pk_cols = sa_inspect(model).mapper.primary_key
        return {col.name for col in pk_cols}
    except Exception:
        return {"id"}


def _derive_flat_schemas(
    model: type[SQLModel],
    *,
    include: set[str] | None = None,
    exclude: set[str] | None = None,
    exclude_pk_from_create: bool = True,
    nested_writes: NestedWriteConfig | None = None,
) -> SchemaConfig:
    pk_fields = _get_pk_field_names(model)
    all_fields = model.model_fields
    exclude_set = exclude or set()

    read_fields: dict[str, ModelFieldDefinition] = {}
    create_fields: dict[str, ModelFieldDefinition] = {}
    update_fields: dict[str, ModelFieldDefinition] = {}

    for name, field_info in all_fields.items():
        if include and name not in include:
            continue
        if name in exclude_set:
            continue

        annotation = field_info.annotation
        default = field_info.default
        if field_info.is_required():
            default = ...

        read_fields[name] = (annotation, default)

        if not (exclude_pk_from_create and name in pk_fields):
            create_fields[name] = (annotation, default)

        if name not in pk_fields:
            opt_annotation = annotation
            opt_default = default
            if opt_default is ...:
                opt_annotation = annotation | None
                opt_default = None
            update_fields[name] = (opt_annotation, opt_default)

    # Backward compatibility: explicit nested relation schema fields in flat models.
    if nested_writes:
        for rel_name, relation_cfg in nested_writes.fields.items():
            related = _derive_flat_schemas(relation_cfg.model)
            create_fields[rel_name] = cast(
                ModelFieldDefinition,
                (related.create | None, None),
            )
            update_fields[rel_name] = cast(
                ModelFieldDefinition,
                (related.update | None, None),
            )
            read_fields[rel_name] = cast(
                ModelFieldDefinition,
                (related.read | None, None),
            )

    _create_model = cast(Callable[..., type[BaseModel]], create_model)
    model_name = model.__name__
    read_schema = _create_model(f"{model_name}Read", __base__=BaseModel, **read_fields)
    create_schema = _create_model(
        f"{model_name}Create", __base__=_StrictInputModel, **create_fields
    )
    update_schema = _create_model(
        f"{model_name}Update", __base__=_StrictInputModel, **update_fields
    )
    return SchemaConfig(create=create_schema, read=read_schema, update=update_schema)


def _augment_update_with_optional_pk(
    model: type[SQLModel], update_schema: type[BaseModel]
) -> type[BaseModel]:
    try:
        pk_name, pk_type = resolve_pk_field(model, None)
    except ValueError:
        return update_schema
    if pk_name in update_schema.model_fields:
        return update_schema
    _create_model = cast(Callable[..., type[BaseModel]], create_model)
    return _create_model(
        f"{model.__name__}NestedNodeUpdate",
        __base__=update_schema,
        **{pk_name: (pk_type | None, None)},
    )


def _resolve_relation_map(
    model: type[SQLModel],
    nested_writes: NestedWriteConfig | None,
    *,
    auto_discover_relations: bool,
) -> dict[str, tuple[type[SQLModel], RelationInfo | None]]:
    relation_map: dict[str, tuple[type[SQLModel], RelationInfo | None]] = {}

    if auto_discover_relations:
        for name, rel_info in inspect_relations(model).items():
            relation_map[name] = (rel_info.target_model, rel_info)

    if nested_writes:
        for rel_name, relation_cfg in nested_writes.fields.items():
            rel_info = relation_map.get(rel_name, (relation_cfg.model, None))[1]
            relation_map[rel_name] = (relation_cfg.model, rel_info)

    return relation_map


def _build_many_to_many_item_schema(
    *,
    model: type[SQLModel],
    relation_name: str,
    relation_info: RelationInfo,
    related_schema: type[BaseModel],
    flat_cache: dict[type[SQLModel], SchemaConfig],
) -> type[BaseModel]:
    _create_model = cast(Callable[..., type[BaseModel]], create_model)

    link_schema_type: type[BaseModel] | None = None
    if relation_info.link_model is not None:
        if relation_info.link_model not in flat_cache:
            flat_cache[relation_info.link_model] = _derive_flat_schemas(
                relation_info.link_model
            )
        link_schema_type = flat_cache[relation_info.link_model].update

    item_fields: dict[str, ModelFieldDefinition] = {
        "entity": cast(ModelFieldDefinition, (related_schema, ...))
    }
    if link_schema_type is not None:
        item_fields["link"] = cast(ModelFieldDefinition, (link_schema_type | None, None))
    else:
        item_fields["link"] = cast(ModelFieldDefinition, (dict[str, object] | None, None))

    return _create_model(
        f"{model.__name__}{relation_name.capitalize()}NestedItem",
        __base__=_StrictInputModel,
        **item_fields,
    )


def _build_nested_update_schema(
    *,
    model: type[SQLModel],
    base_update_schema: type[BaseModel],
    auto_discover_relations: bool,
    nested_writes: NestedWriteConfig | None,
    active_path: tuple[type[SQLModel], ...],
    nested_cache: dict[
        tuple[type[SQLModel], tuple[type[SQLModel], ...]],
        type[BaseModel],
    ],
    flat_cache: dict[type[SQLModel], SchemaConfig],
) -> type[BaseModel]:
    cache_key = (model, active_path)
    if cache_key in nested_cache:
        return nested_cache[cache_key]

    relation_map = _resolve_relation_map(
        model,
        nested_writes,
        auto_discover_relations=auto_discover_relations,
    )

    nested_fields: dict[str, ModelFieldDefinition] = {}
    for relation_name, (related_model, rel_info) in relation_map.items():
        if related_model not in flat_cache:
            flat_cache[related_model] = _derive_flat_schemas(related_model)
        related_base_update = _augment_update_with_optional_pk(
            related_model, flat_cache[related_model].update
        )
        if related_model in active_path:
            related_nested_schema = related_base_update
        else:
            related_nested_schema = _build_nested_update_schema(
                model=related_model,
                base_update_schema=related_base_update,
                auto_discover_relations=auto_discover_relations,
                nested_writes=None,
                active_path=(*active_path, model),
                nested_cache=nested_cache,
                flat_cache=flat_cache,
            )

        if rel_info is not None and rel_info.is_many_to_many:
            item_schema = _build_many_to_many_item_schema(
                model=model,
                relation_name=relation_name,
                relation_info=rel_info,
                related_schema=related_nested_schema,
                flat_cache=flat_cache,
            )
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (GenericAlias(list, item_schema) | None, None),
            )
        elif rel_info is not None and rel_info.is_to_many:
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (GenericAlias(list, related_nested_schema) | None, None),
            )
        else:
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (related_nested_schema | None, None),
            )

    if not nested_fields:
        nested_cache[cache_key] = base_update_schema
        return base_update_schema

    _create_model = cast(Callable[..., type[BaseModel]], create_model)
    nested_schema = _create_model(
        f"{model.__name__}NestedUpdate",
        __base__=base_update_schema,
        **nested_fields,
    )
    nested_cache[cache_key] = nested_schema
    return nested_schema


def _build_nested_create_schema(
    *,
    model: type[SQLModel],
    base_schema: type[BaseModel],
    auto_discover_relations: bool,
    nested_writes: NestedWriteConfig | None,
    active_path: tuple[type[SQLModel], ...],
    nested_cache: dict[
        tuple[type[SQLModel], tuple[type[SQLModel], ...]],
        type[BaseModel],
    ],
    flat_cache: dict[type[SQLModel], SchemaConfig],
) -> type[BaseModel]:
    cache_key = (model, active_path)
    if cache_key in nested_cache:
        return nested_cache[cache_key]

    relation_map = _resolve_relation_map(
        model,
        nested_writes,
        auto_discover_relations=auto_discover_relations,
    )

    nested_fields: dict[str, ModelFieldDefinition] = {}
    for relation_name, (related_model, rel_info) in relation_map.items():
        if related_model not in flat_cache:
            flat_cache[related_model] = _derive_flat_schemas(related_model)
        related_base_schema = _augment_update_with_optional_pk(
            related_model, flat_cache[related_model].update
        )
        if related_model in active_path:
            related_nested_schema = related_base_schema
        else:
            related_nested_schema = _build_nested_create_schema(
                model=related_model,
                base_schema=related_base_schema,
                auto_discover_relations=auto_discover_relations,
                nested_writes=None,
                active_path=(*active_path, model),
                nested_cache=nested_cache,
                flat_cache=flat_cache,
            )

        if rel_info is not None and rel_info.is_many_to_many:
            item_schema = _build_many_to_many_item_schema(
                model=model,
                relation_name=relation_name,
                relation_info=rel_info,
                related_schema=related_nested_schema,
                flat_cache=flat_cache,
            )
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (GenericAlias(list, item_schema) | None, None),
            )
        elif rel_info is not None and rel_info.is_to_many:
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (GenericAlias(list, related_nested_schema) | None, None),
            )
        else:
            nested_fields[relation_name] = cast(
                ModelFieldDefinition,
                (related_nested_schema | None, None),
            )

    if not nested_fields:
        nested_cache[cache_key] = base_schema
        return base_schema

    _create_model = cast(Callable[..., type[BaseModel]], create_model)
    nested_schema = _create_model(
        f"{model.__name__}NestedCreate",
        __base__=base_schema,
        **nested_fields,
    )
    nested_cache[cache_key] = nested_schema
    return nested_schema


def _derive_schema_container_impl(
    model: type[SQLModel],
    *,
    include: set[str] | None,
    exclude: set[str] | None,
    exclude_pk_from_create: bool,
    nested_writes: NestedWriteConfig | None,
    auto_discover_relations: bool,
) -> SchemaContainer:
    flat = _derive_flat_schemas(
        model,
        include=include,
        exclude=exclude,
        exclude_pk_from_create=exclude_pk_from_create,
        nested_writes=nested_writes,
    )
    nested_create = _build_nested_create_schema(
        model=model,
        base_schema=flat.create,
        auto_discover_relations=auto_discover_relations,
        nested_writes=nested_writes,
        active_path=(),
        nested_cache={},
        flat_cache={model: flat},
    )
    nested_update = _build_nested_update_schema(
        model=model,
        base_update_schema=flat.update,
        auto_discover_relations=auto_discover_relations,
        nested_writes=nested_writes,
        active_path=(),
        nested_cache={},
        flat_cache={model: flat},
    )
    return SchemaContainer(
        create=flat.create,
        read=flat.read,
        update=flat.update,
        nested_create=nested_create,
        nested_update=nested_update,
    )


@lru_cache(maxsize=256)
def _derive_schema_container_cached(
    model: type[SQLModel],
    include: frozenset[str] | None,
    exclude: frozenset[str],
    exclude_pk_from_create: bool,
    auto_discover_relations: bool,
) -> SchemaContainer:
    include_set = set(include) if include is not None else None
    exclude_set = set(exclude)
    return _derive_schema_container_impl(
        model,
        include=include_set,
        exclude=exclude_set,
        exclude_pk_from_create=exclude_pk_from_create,
        nested_writes=None,
        auto_discover_relations=auto_discover_relations,
    )


def derive_schema_container(
    model: type[SQLModel],
    *,
    include: set[str] | None = None,
    exclude: set[str] | None = None,
    exclude_pk_from_create: bool = True,
    nested_writes: NestedWriteConfig | None = None,
    auto_discover_relations: bool = True,
) -> SchemaContainer:
    """Derive Create/Read/Update/NestedCreate/NestedUpdate schemas from a SQLModel."""
    include_key = frozenset(include) if include is not None else None
    exclude_key = frozenset(exclude or set())
    if nested_writes is None:
        return _derive_schema_container_cached(
            model,
            include_key,
            exclude_key,
            exclude_pk_from_create,
            auto_discover_relations,
        )
    return _derive_schema_container_impl(
        model,
        include=set(include) if include is not None else None,
        exclude=set(exclude or set()),
        exclude_pk_from_create=exclude_pk_from_create,
        nested_writes=nested_writes,
        auto_discover_relations=auto_discover_relations,
    )


def derive_schemas(
    model: type[SQLModel],
    *,
    include: set[str] | None = None,
    exclude: set[str] | None = None,
    exclude_pk_from_create: bool = True,
    nested_writes: NestedWriteConfig | None = None,
) -> SchemaConfig:
    """Derive Create/Read/Update/NestedCreate/NestedUpdate schemas from a SQLModel."""
    return derive_schema_container(
        model,
        include=include,
        exclude=exclude,
        exclude_pk_from_create=exclude_pk_from_create,
        nested_writes=nested_writes,
    ).to_config()



"""CRUD router generation from SQLModel models."""

from collections.abc import Callable, Collection, Sequence
from enum import Enum
from typing import cast

from fastapi import APIRouter
from sqlmodel import SQLModel
from typing_extensions import Self, TypedDict

from auen._async_endpoints import (
    add_async_bulk_create,
    add_async_bulk_delete,
    add_async_bulk_update,
    add_async_create,
    add_async_delete,
    add_async_list,
    add_async_nested_create,
    add_async_nested_update,
    add_async_read,
    add_async_update,
)
from auen._endpoints import RouterContext
from auen._query import resolve_pk_field
from auen.config import (
    ALL_OPERATIONS,
    AuthConfig,
    FilterConfig,
    HooksConfig,
    NestedWriteConfig,
    Operation,
    PaginationConfig,
    SchemaConfig,
)
from auen.policy import AllowAll, CrudPolicy
from auen.repository import AsyncCrudRepository, AsyncSqlModelRepository
from auen.schemas import derive_schema_container
from auen.types import PKType, SessionDep


class _BuildKwargs(TypedDict):
    schemas: SchemaConfig | None
    prefix: str
    tags: list[str | Enum]
    operations: Collection[Operation]
    auth: AuthConfig | None
    policy: CrudPolicy | None
    id_field: str
    pagination: PaginationConfig | None
    filters: FilterConfig | None
    repository_factory: Callable[[type[SQLModel]], AsyncCrudRepository]
    hooks: HooksConfig | None
    nested_writes: NestedWriteConfig | None
    nested_policy_registry: dict[type[SQLModel], CrudPolicy]
    include_in_schema: set[Operation] | None
    operation_id_prefix: str


_OP_BUILDERS = {
    Operation.CREATE: add_async_create,
    Operation.NESTED_CREATE: add_async_nested_create,
    Operation.LIST: add_async_list,
    Operation.READ: add_async_read,
    Operation.UPDATE: add_async_update,
    Operation.NESTED_UPDATE: add_async_nested_update,
    Operation.DELETE: add_async_delete,
    Operation.BULK_CREATE: add_async_bulk_create,
    Operation.BULK_UPDATE: add_async_bulk_update,
    Operation.BULK_DELETE: add_async_bulk_delete,
}


def _build_one_router(
    *,
    model: type[SQLModel],
    session_dep: Callable[..., SessionDep],
    schemas: SchemaConfig | None,
    prefix: str,
    tags: list[str | Enum],
    operations: Collection[Operation],
    auth: AuthConfig | None,
    policy: CrudPolicy | None,
    id_field: str,
    pagination: PaginationConfig | None,
    filters: FilterConfig | None,
    repository_factory: Callable[[type[SQLModel]], AsyncCrudRepository],
    hooks: HooksConfig | None,
    nested_writes: NestedWriteConfig | None,
    nested_policy_registry: dict[type[SQLModel], CrudPolicy],
    include_in_schema: set[Operation] | None,
    operation_id_prefix: str,
) -> APIRouter:
    """Build a single APIRouter with CRUD endpoints for one model."""
    model_name = model.__name__
    pk_name, pk_type = resolve_pk_field(model, id_field or None)
    pk_type = cast("PKType", pk_type)

    if schemas is None:
        schemas = derive_schema_container(model).to_config()
    if not prefix:
        prefix = f"/{model_name.lower()}s"
    elif not prefix.startswith("/"):
        prefix = f"/{prefix}"
    tags_list: list[str | Enum] = [model_name] if not tags else list(tags)

    router = APIRouter(prefix=prefix, tags=tags_list)

    effective_filters = (
        filters if filters and (filters.fields or filters.sort_fields) else None
    )

    ctx = RouterContext(
        router=router,
        model=model,
        model_name=model_name,
        schemas=schemas,
        policy=policy or AllowAll(),
        session_dep=session_dep,
        auth=auth,
        pk_name=pk_name,
        pk_type=pk_type,
        pagination=pagination or PaginationConfig(),
        filters=effective_filters,
        repository_factory=repository_factory,
        hooks=hooks or HooksConfig(),
        nested_writes=nested_writes,
        nested_policy_registry=nested_policy_registry,
        include_in_schema=include_in_schema or set(operations),
        operation_id_prefix=operation_id_prefix or model_name.lower(),
    )

    for op, builder in _OP_BUILDERS.items():
        if op in operations:
            builder(ctx)

    return router


class CrudRouterBuilder:
    """Fluent builder for CRUD routers.

    Accepts a single model or a list of models.
    Use ``build()`` for a single model, ``build_all()`` for multiple.
    """

    def __init__(
        self,
        models: type[SQLModel] | list[type[SQLModel]],
        session_dep: Callable[..., SessionDep],
    ) -> None:
        if isinstance(models, list):
            self._models = models
            self._single = False
        else:
            self._models = [models]
            self._single = True
        self._session_dep = session_dep
        self._schemas: SchemaConfig | None = None
        self._prefix: str = ""
        self._tags: list[str | Enum] = []
        self._operations: Collection[Operation] = ALL_OPERATIONS
        self._auth: AuthConfig | None = None
        self._policy: CrudPolicy | None = None
        self._id_field: str = ""
        self._pagination: PaginationConfig | None = None
        self._filters: FilterConfig | None = None
        self._repository_factory: Callable[[type[SQLModel]], AsyncCrudRepository] = (
            AsyncSqlModelRepository
        )
        self._hooks: HooksConfig | None = None
        self._nested_writes: NestedWriteConfig | None = None
        self._nested_policy_registry: dict[type[SQLModel], CrudPolicy] = {}
        self._include_in_schema: set[Operation] | None = None
        self._operation_id_prefix: str = ""

    @classmethod
    def for_model(
        cls,
        model: type[SQLModel] | list[type[SQLModel]],
        session_dep: Callable[..., SessionDep],
    ) -> Self:
        return cls(model, session_dep)

    # --- Fluent setters ---

    def with_schemas(self, schemas: SchemaConfig) -> Self:
        self._schemas = schemas
        return self

    def with_prefix(self, prefix: str) -> Self:
        self._prefix = prefix
        return self

    def with_tags(self, tags: Sequence[str | Enum]) -> Self:
        self._tags = list(tags)
        return self

    def with_operations(self, operations: Collection[Operation]) -> Self:
        self._operations = operations
        return self

    def with_auth(self, auth: AuthConfig) -> Self:
        self._auth = auth
        return self

    def with_policy(self, policy: CrudPolicy) -> Self:
        self._policy = policy
        return self

    def with_id_field(self, id_field: str) -> Self:
        self._id_field = id_field
        return self

    def with_pagination(self, pagination: PaginationConfig) -> Self:
        self._pagination = pagination
        return self

    def with_filters(self, filters: FilterConfig) -> Self:
        self._filters = filters
        return self

    def with_repository_factory(
        self,
        repository_factory: Callable[[type[SQLModel]], AsyncCrudRepository],
    ) -> Self:
        self._repository_factory = repository_factory
        return self

    def with_hooks(self, hooks: HooksConfig) -> Self:
        self._hooks = hooks
        return self

    def with_nested_writes(self, nested_writes: NestedWriteConfig) -> Self:
        self._nested_writes = nested_writes
        return self

    def with_nested_policy_registry(
        self, registry: dict[type[SQLModel], CrudPolicy]
    ) -> Self:
        self._nested_policy_registry = registry
        return self

    def with_include_in_schema(self, ops: set[Operation]) -> Self:
        self._include_in_schema = ops
        return self

    def with_operation_id_prefix(self, prefix: str) -> Self:
        self._operation_id_prefix = prefix
        return self

    # --- Build ---

    def _build_kwargs(self) -> _BuildKwargs:
        return {
            "schemas": self._schemas,
            "prefix": self._prefix,
            "tags": self._tags,
            "operations": self._operations,
            "auth": self._auth,
            "policy": self._policy,
            "id_field": self._id_field,
            "pagination": self._pagination,
            "filters": self._filters,
            "repository_factory": self._repository_factory,
            "hooks": self._hooks,
            "nested_writes": self._nested_writes,
            "nested_policy_registry": self._nested_policy_registry,
            "include_in_schema": self._include_in_schema,
            "operation_id_prefix": self._operation_id_prefix,
        }

    def build(self) -> APIRouter:
        """Build a single APIRouter. Raises if multiple models were given."""
        if not self._single:
            msg = "build() requires a single model; use build_all() for multiple models"
            raise ValueError(msg)
        return _build_one_router(
            model=self._models[0],
            session_dep=self._session_dep,
            **self._build_kwargs(),
        )

    def build_all(self) -> list[APIRouter]:
        """Build one APIRouter per model."""
        kwargs = self._build_kwargs()
        return [
            _build_one_router(
                model=model,
                session_dep=self._session_dep,
                **kwargs,
            )
            for model in self._models
        ]



"""Exception type tests."""

from auen.exceptions import ConflictError


def test_conflict_error_default() -> None:
    err = ConflictError()
    assert err.status_code == 409
    assert err.detail == "Resource conflict"



"""Contract tests: verify routes exist and CRUD roundtrip works."""

from __future__ import annotations

from httpx import AsyncClient
from tests.conftest import Book, SessionFactory

from auen import CrudRouterBuilder, Operation


async def test_routes_exist(client: AsyncClient) -> None:
    """All five CRUD routes should be registered."""
    resp = await client.get("/books/")
    assert resp.status_code == 200

    resp = await client.post(
        "/books/",
        json={"title": "Clean Code", "isbn": "978-0132350884"},
    )
    assert resp.status_code == 201


async def test_crud_roundtrip(client: AsyncClient) -> None:
    """Create -> Read -> List -> Update -> Delete roundtrip."""
    # Create
    resp = await client.post(
        "/books/",
        json={
            "title": "Design Patterns",
            "isbn": "978-0201633610",
            "pages": 395,
        },
    )
    assert resp.status_code == 201
    book = resp.json()
    book_id = book["id"]
    assert book["title"] == "Design Patterns"

    # Read
    resp = await client.get(f"/books/{book_id}")
    assert resp.status_code == 200
    assert resp.json()["isbn"] == "978-0201633610"

    # List
    resp = await client.get("/books/")
    assert resp.status_code == 200
    books = resp.json()
    assert any(b["id"] == book_id for b in books)

    # Update (partial - exclude_unset)
    resp = await client.patch(f"/books/{book_id}", json={"pages": 416})
    assert resp.status_code == 200
    updated = resp.json()
    assert updated["pages"] == 416
    assert updated["title"] == "Design Patterns"  # not overwritten

    # Delete
    resp = await client.delete(f"/books/{book_id}")
    assert resp.status_code == 200

    # Verify 404
    resp = await client.get(f"/books/{book_id}")
    assert resp.status_code == 404


async def test_not_found(client: AsyncClient) -> None:
    resp = await client.get("/books/99999")
    assert resp.status_code == 404


async def test_subset_operations(get_session: SessionFactory) -> None:
    """Only requested operations should be generated."""
    from fastapi import FastAPI
    from httpx import ASGITransport, AsyncClient

    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_operations({Operation.LIST, Operation.READ})
        .build()
    )
    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        resp = await c.get("/books/")
        assert resp.status_code == 200

        resp = await c.post(
            "/books/",
            json={
                "title": "X",
                "isbn": "Y",
            },
        )
        assert resp.status_code == 405



"""Tests for lifecycle hooks."""

from __future__ import annotations

from collections.abc import AsyncGenerator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from pydantic import BaseModel
from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession
from tests.conftest import Book, SessionFactory

from auen import CrudRouterBuilder, HooksConfig, Operation
from auen.types import User


@pytest.fixture
def hook_log() -> list[tuple[str, int | str | None]]:
    return []


@pytest.fixture
def hooks_app(
    get_session: SessionFactory, hook_log: list[tuple[str, int | str | None]]
) -> FastAPI:
    async def before_create(session: AsyncSession, obj_in: BaseModel, user: User) -> None:
        hook_log.append(("before_create", type(obj_in).__name__))

    async def after_create(session: AsyncSession, db_obj: Book, user: User) -> None:
        hook_log.append(("after_create", db_obj.id))

    async def before_update(session: AsyncSession, db_obj: Book, user: User) -> None:
        hook_log.append(("before_update", db_obj.id))

    async def after_update(session: AsyncSession, db_obj: Book, user: User) -> None:
        hook_log.append(("after_update", db_obj.title))

    async def before_delete(session: AsyncSession, db_obj: Book, user: User) -> None:
        hook_log.append(("before_delete", db_obj.id))

    async def after_delete(session: AsyncSession, db_obj: Book, user: User) -> None:
        hook_log.append(("after_delete", db_obj.id))

    from tests.conftest import Book

    application = FastAPI()
    application.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(
            HooksConfig(
                before_create=before_create,
                after_create=after_create,
                before_update=before_update,
                after_update=after_update,
                before_delete=before_delete,
                after_delete=after_delete,
            )
        )
        .build()
    )
    return application


@pytest.fixture
async def hooks_client(
    hooks_app: FastAPI,
) -> AsyncGenerator[AsyncClient, None]:
    transport = ASGITransport(app=hooks_app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        yield c


async def test_create_hooks_fire(
    hooks_client: AsyncClient, hook_log: list[tuple[str, int | str | None]]
) -> None:
    resp = await hooks_client.post("/books/", json={"title": "Test", "isbn": "123"})
    assert resp.status_code == 201
    assert ("before_create", "BookCreate") in hook_log
    assert ("after_create", resp.json()["id"]) in hook_log


async def test_update_hooks_fire(
    hooks_client: AsyncClient, hook_log: list[tuple[str, int | str | None]]
) -> None:
    resp = await hooks_client.post("/books/", json={"title": "Old", "isbn": "456"})
    book_id = resp.json()["id"]
    resp = await hooks_client.patch(f"/books/{book_id}", json={"title": "New"})
    assert resp.status_code == 200
    assert ("before_update", book_id) in hook_log
    assert ("after_update", "New") in hook_log


async def test_delete_hooks_fire(
    hooks_client: AsyncClient, hook_log: list[tuple[str, int | str | None]]
) -> None:
    resp = await hooks_client.post("/books/", json={"title": "Bye", "isbn": "789"})
    book_id = resp.json()["id"]
    resp = await hooks_client.delete(f"/books/{book_id}")
    assert resp.status_code == 200
    assert ("before_delete", book_id) in hook_log
    assert ("after_delete", book_id) in hook_log


async def test_hooks_before_after_create(get_session: SessionFactory) -> None:
    called = []

    async def before_create(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before_create")

    async def after_create(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after_create")

    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(before_create=before_create, after_create=after_create))
        .with_operations({Operation.CREATE})
        .build()
    )
    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        resp = await c.post("/books/", json={"title": "T", "isbn": "I"})
        assert resp.status_code == 201
    assert called == ["before_create", "after_create"]


async def test_hooks_before_after_update(get_session: SessionFactory) -> None:
    called = []

    async def before_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before_update")

    async def after_update(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after_update")

    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(before_update=before_update, after_update=after_update))
        .with_operations({Operation.CREATE, Operation.UPDATE})
        .build()
    )
    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        resp = await c.post("/books/", json={"title": "T", "isbn": "I"})
        book_id = resp.json()["id"]
        resp = await c.patch(f"/books/{book_id}", json={"title": "Updated"})
        assert resp.status_code == 200
    assert called == ["before_update", "after_update"]


async def test_hooks_before_after_delete(get_session: SessionFactory) -> None:
    called = []

    async def before_delete(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("before_delete")

    async def after_delete(session: AsyncSession, obj: SQLModel, user: User) -> None:
        called.append("after_delete")

    app = FastAPI()
    app.include_router(
        CrudRouterBuilder.for_model(Book, get_session)
        .with_hooks(HooksConfig(before_delete=before_delete, after_delete=after_delete))
        .with_operations({Operation.CREATE, Operation.DELETE})
        .build()
    )
    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as c:
        resp = await c.post("/books/", json={"title": "T", "isbn": "I"})
        book_id = resp.json()["id"]
        resp = await c.delete(f"/books/{book_id}")
        assert resp.status_code == 200
    assert called == ["before_delete", "after_delete"]



"""Simple test for async support."""

from __future__ import annotations

from collections.abc import AsyncGenerator, Callable

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlmodel import Field, SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

from auen import CrudRouterBuilder, derive_schemas


class AsyncItem(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    name: str


@pytest.fixture
async def async_engine() -> AsyncEngine:
    eng = create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        echo=False,
    )
    async with eng.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)
    return eng


@pytest.fixture
async def async_get_session(
    async_engine: AsyncEngine,
) -> Callable[[], AsyncGenerator[AsyncSession, None]]:
    async def _get_session() -> AsyncGenerator[AsyncSession, None]:
        async with AsyncSession(async_engine) as session:
            yield session

    return _get_session


@pytest.fixture
def async_app(
    async_get_session: Callable[[], AsyncGenerator[AsyncSession, None]],
) -> FastAPI:
    """Async app with async endpoints."""
    application = FastAPI()

    application.include_router(
        CrudRouterBuilder.for_model(AsyncItem, async_get_session)
        .with_schemas(derive_schemas(AsyncItem))
        .build()
    )
    return application


@pytest.mark.asyncio
async def test_async_create(async_app: FastAPI) -> None:
    """Test async CREATE operation."""
    transport = ASGITransport(app=async_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post("/asyncitems/", json={"name": "test"})
        assert resp.status_code == 201
        item = resp.json()
        assert item["name"] == "test"
        assert "id" in item


@pytest.mark.asyncio
async def test_async_read(async_app: FastAPI) -> None:
    """Test async READ operation."""
    transport = ASGITransport(app=async_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        # Create an item first
        resp = await client.post("/asyncitems/", json={"name": "test"})
        item_id = resp.json()["id"]

        # Read it back
        resp = await client.get(f"/asyncitems/{item_id}")
        assert resp.status_code == 200
        assert resp.json()["name"] == "test"


@pytest.mark.asyncio
async def test_async_list(async_app: FastAPI) -> None:
    """Test async LIST operation."""
    transport = ASGITransport(app=async_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        # Create multiple items
        for i in range(3):
            await client.post("/asyncitems/", json={"name": f"item{i}"})

        # List them
        resp = await client.get("/asyncitems/")
        assert resp.status_code == 200
        items = resp.json()
        assert len(items) == 3
        assert all("id" in item and "name" in item for item in items)



"""OpenAPI snapshot test."""

from __future__ import annotations

import json
from pathlib import Path

import pytest
from fastapi import FastAPI

SNAPSHOT_DIR = Path(__file__).parent / "snapshots"


def test_openapi_snapshot(app: FastAPI) -> None:
    """OpenAPI schema should be stable."""
    schema = app.openapi()
    snapshot_path = SNAPSHOT_DIR / "openapi.json"

    if not snapshot_path.exists():
        SNAPSHOT_DIR.mkdir(parents=True, exist_ok=True)
        snapshot_path.write_text(json.dumps(schema, indent=2, sort_keys=True))
        pytest.skip("Snapshot created; re-run to verify.")

    expected = json.loads(snapshot_path.read_text())
    # Compare paths and schemas (stable subset)
    assert set(schema["paths"].keys()) == set(expected["paths"].keys())
    assert set(schema.get("components", {}).get("schemas", {}).keys()) == set(
        expected.get("components", {}).get("schemas", {}).keys()
    )



"""Unit coverage for nested mutation helpers."""

from __future__ import annotations

from types import SimpleNamespace
from typing import cast
from unittest.mock import AsyncMock

import pytest
from pydantic import BaseModel, ConfigDict
from sqlalchemy.orm import RelationshipDirection
from sqlmodel import Field, SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

from auen import CrudPolicy
from auen._nested_mutations import (
    _apply_many_to_many,
    _apply_one_to_many,
    _apply_relation,
    _as_list,
    _create_model_graph,
    _many_to_many_pairs,
    _many_to_one_fk_field,
    _MutationContext,
    _one_to_many_pair,
    _update_model_graph,
)
from auen._relations import RelationInfo
from auen.exceptions import ConfigurationError, ForbiddenError, NotFoundError


class AnyCreate(BaseModel):
    model_config = ConfigDict(extra="allow")


class AnyUpdate(BaseModel):
    model_config = ConfigDict(extra="allow")


class UnitParent(SQLModel, table=True):
    __tablename__ = "unit_parent"
    id: int | None = Field(default=None, primary_key=True)
    parent_fk: int | None = None


class UnitChild(SQLModel, table=True):
    __tablename__ = "unit_child"
    id: int | None = Field(default=None, primary_key=True)
    parent_id: int | None = Field(default=None, foreign_key="unit_parent.id")
    name: str | None = None


class UnitTag(SQLModel, table=True):
    __tablename__ = "unit_tag"
    id: int | None = Field(default=None, primary_key=True)
    name: str | None = None


class UnitLink(SQLModel, table=True):
    __tablename__ = "unit_link"
    parent_id: int | None = Field(
        default=None,
        foreign_key="unit_parent.id",
        primary_key=True,
    )
    tag_id: int | None = Field(
        default=None,
        foreign_key="unit_tag.id",
        primary_key=True,
    )
    weight: int | None = None


def _ctx(session: AsyncSession | object) -> _MutationContext:
    return _MutationContext(
        session=cast(AsyncSession, session),
        user=None,
        nested_policy_registry={},
        relation_policy_overrides={},
    )


def _rel(
    *,
    name: str,
    direction: RelationshipDirection,
    uselist: bool,
    local_columns: tuple[str, ...] = ("parent_fk",),
    sync_pairs: tuple[tuple[str, str], ...] = (("id", "parent_id"),),
    secondary_sync_pairs: tuple[tuple[str, str], ...] = (("id", "tag_id"),),
    target_model: type[SQLModel] = UnitChild,
    link_model: type[SQLModel] | None = None,
) -> RelationInfo:
    return RelationInfo(
        name=name,
        parent_model=UnitParent,
        target_model=target_model,
        direction=direction,
        uselist=uselist,
        secondary=None,
        local_columns=local_columns,
        remote_columns=(),
        sync_pairs=sync_pairs,
        secondary_sync_pairs=secondary_sync_pairs,
        local_remote_pairs=(),
        link_model=link_model,
    )


def test_pair_helpers_and_list_validation_errors() -> None:
    with pytest.raises(Exception, match="must be a list"):
        _as_list("x", relation_name="rel")

    with pytest.raises(ConfigurationError, match="exactly one local FK"):
        _many_to_one_fk_field(
            _rel(
                name="rel",
                direction=RelationshipDirection.MANYTOONE,
                uselist=False,
                local_columns=("a", "b"),
            )
        )

    with pytest.raises(ConfigurationError, match="exactly one sync pair"):
        _one_to_many_pair(
            _rel(
                name="children",
                direction=RelationshipDirection.ONETOMANY,
                uselist=True,
                sync_pairs=(("id", "parent_id"), ("id", "parent_id2")),
            )
        )

    with pytest.raises(ConfigurationError, match="exactly one parent and one target"):
        _many_to_many_pairs(
            _rel(
                name="tags",
                direction=RelationshipDirection.MANYTOMANY,
                uselist=True,
                sync_pairs=(("id", "parent_id"), ("id", "parent_id2")),
            )
        )


@pytest.mark.asyncio
async def test_create_model_graph_processes_m2o_and_followup_relations(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = SimpleNamespace(add=lambda _obj: None, flush=AsyncMock())
    ctx = _ctx(session)

    relation_data = {"author": {"id": 1}, "chapters": [{"name": "c1"}]}
    relations = {
        "author": _rel(
            name="author",
            direction=RelationshipDirection.MANYTOONE,
            uselist=False,
            local_columns=("parent_fk",),
            target_model=UnitChild,
        ),
        "chapters": _rel(
            name="chapters",
            direction=RelationshipDirection.ONETOMANY,
            uselist=True,
            sync_pairs=(("id", "parent_id"),),
            target_model=UnitChild,
        ),
    }
    monkeypatch.setattr(
        "auen._nested_mutations._split_payload",
        lambda _m, _p: ({"parent_fk": 1}, dict(relation_data), relations),
    )
    monkeypatch.setattr(
        "auen._nested_mutations._flat_schemas",
        lambda _m: (AnyCreate, AnyUpdate),
    )
    m2o = AsyncMock()
    followup = AsyncMock()
    monkeypatch.setattr("auen._nested_mutations._apply_many_to_one", m2o)
    monkeypatch.setattr("auen._nested_mutations._apply_relation", followup)

    obj = await _create_model_graph(ctx, UnitParent, {"x": 1})
    assert isinstance(obj, UnitParent)
    assert m2o.await_count == 1
    assert followup.await_count == 1


@pytest.mark.asyncio
async def test_update_model_graph_processes_relation_loop(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = SimpleNamespace(add=lambda _obj: None, flush=AsyncMock())
    ctx = _ctx(session)
    relations = {
        "author": _rel(
            name="author",
            direction=RelationshipDirection.MANYTOONE,
            uselist=False,
            local_columns=("parent_fk",),
            target_model=UnitChild,
        )
    }
    monkeypatch.setattr(
        "auen._nested_mutations._split_payload",
        lambda _m, _p: ({"parent_fk": 1}, {"author": {"id": 1}}, relations),
    )
    monkeypatch.setattr(
        "auen._nested_mutations._flat_schemas",
        lambda _m: (AnyCreate, AnyUpdate),
    )
    apply_relation = AsyncMock()
    monkeypatch.setattr("auen._nested_mutations._apply_relation", apply_relation)

    await _update_model_graph(ctx, UnitParent, UnitParent(id=1), {"x": 1})
    assert apply_relation.await_count == 1


@pytest.mark.asyncio
async def test_apply_relation_requires_parent_for_to_many() -> None:
    ctx = _ctx(SimpleNamespace())
    scalar_data: dict[str, object] = {}
    with pytest.raises(Exception, match="persisted parent object"):
        await _apply_relation(
            ctx,
            parent_model=UnitParent,
            parent_obj=None,
            relation_name="tags",
            rel_info=_rel(
                name="tags",
                direction=RelationshipDirection.MANYTOMANY,
                uselist=True,
                target_model=UnitTag,
            ),
            relation_payload=[],
            parent_scalar_data=scalar_data,
        )
    with pytest.raises(Exception, match="persisted parent object"):
        await _apply_relation(
            ctx,
            parent_model=UnitParent,
            parent_obj=None,
            relation_name="children",
            rel_info=_rel(
                name="children",
                direction=RelationshipDirection.ONETOMANY,
                uselist=True,
                target_model=UnitChild,
            ),
            relation_payload=[],
            parent_scalar_data=scalar_data,
        )
    with pytest.raises(Exception, match="persisted parent object"):
        await _apply_relation(
            ctx,
            parent_model=UnitParent,
            parent_obj=None,
            relation_name="child",
            rel_info=_rel(
                name="child",
                direction=RelationshipDirection.ONETOMANY,
                uselist=False,
                target_model=UnitChild,
            ),
            relation_payload={},
            parent_scalar_data=scalar_data,
        )

    with pytest.raises(Exception, match="Unsupported nested relation type"):
        await _apply_relation(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="weird",
            rel_info=_rel(
                name="weird",
                direction=cast(RelationshipDirection, "UNKNOWN"),
                uselist=False,
                target_model=UnitChild,
            ),
            relation_payload={},
            parent_scalar_data=scalar_data,
        )


@pytest.mark.asyncio
async def test_apply_one_to_many_parent_flush_and_error_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    parent = UnitParent(id=None)

    async def _flush_assign() -> None:
        parent.id = 99

    session = SimpleNamespace(flush=_flush_assign, get=AsyncMock(return_value=None))
    ctx = _ctx(session)
    monkeypatch.setattr("auen._nested_mutations._create_model_graph", AsyncMock())
    monkeypatch.setattr("auen._nested_mutations._update_model_graph", AsyncMock())

    rel = _rel(
        name="children",
        direction=RelationshipDirection.ONETOMANY,
        uselist=True,
        sync_pairs=(("id", "parent_id"),),
        target_model=UnitChild,
    )

    # parent id is assigned during flush, then create path runs and continues
    await _apply_one_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=parent,
        relation_name="children",
        rel_info=rel,
        relation_payload=[{"name": "n1"}],
    )
    assert parent.id == 99

    with pytest.raises(Exception, match="Conflict between 'parent_id'"):
        await _apply_one_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="children",
            rel_info=rel,
            relation_payload=[{"parent_id": 2}],
        )

    with pytest.raises(NotFoundError):
        await _apply_one_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="children",
            rel_info=rel,
            relation_payload=[{"id": 123}],
        )

    session.get = AsyncMock(return_value=UnitChild(id=123, parent_id=7))
    with pytest.raises(Exception, match="is not related"):
        await _apply_one_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="children",
            rel_info=rel,
            relation_payload=[{"id": 123}],
        )

    # existing child update path calls nested updater
    update_child = AsyncMock()
    monkeypatch.setattr("auen._nested_mutations._update_model_graph", update_child)
    session.get = AsyncMock(return_value=UnitChild(id=123, parent_id=1))
    await _apply_one_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=UnitParent(id=1),
        relation_name="children",
        rel_info=rel,
        relation_payload=[{"id": 123, "name": "ok"}],
    )
    assert update_child.await_count == 1


@pytest.mark.asyncio
async def test_apply_many_to_many_error_and_link_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    parent = UnitParent(id=None)

    async def _flush_assign() -> None:
        parent.id = 10

    session = SimpleNamespace(flush=_flush_assign, get=AsyncMock(return_value=None))
    ctx = _ctx(session)
    monkeypatch.setattr("auen._nested_mutations._create_model_graph", AsyncMock())
    monkeypatch.setattr("auen._nested_mutations._update_model_graph", AsyncMock())
    monkeypatch.setattr(
        "auen._nested_mutations._flat_schemas",
        lambda _m: (AnyCreate, AnyUpdate),
    )

    rel_with_link = _rel(
        name="tags",
        direction=RelationshipDirection.MANYTOMANY,
        uselist=True,
        sync_pairs=(("id", "parent_id"),),
        secondary_sync_pairs=(("id", "tag_id"),),
        target_model=UnitTag,
        link_model=UnitLink,
    )

    with pytest.raises(Exception, match="must contain 'entity'"):
        await _apply_many_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=parent,
            relation_name="tags",
            rel_info=rel_with_link,
            relation_payload=[{}],
        )

    with pytest.raises(NotFoundError):
        await _apply_many_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="tags",
            rel_info=rel_with_link,
            relation_payload=[{"entity": {"id": 999}}],
        )

    # link_model is None path appends to relation list
    tag_obj = UnitTag(id=5)
    session.get = AsyncMock(return_value=tag_obj)
    rel_without_link = _rel(
        name="tags",
        direction=RelationshipDirection.MANYTOMANY,
        uselist=True,
        sync_pairs=(("id", "parent_id"),),
        secondary_sync_pairs=(("id", "tag_id"),),
        target_model=UnitTag,
        link_model=None,
    )
    parent_with_list = SimpleNamespace(id=1, tags=[])
    await _apply_many_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=cast(SQLModel, parent_with_list),
        relation_name="tags",
        rel_info=rel_without_link,
        relation_payload=[{"entity": {"id": 5}}],
    )
    assert tag_obj in parent_with_list.tags

    # link create branch
    async def _exec_none(_stmt: object) -> object:
        return SimpleNamespace(first=lambda: None)

    added: list[object] = []
    session.exec = _exec_none
    session.add = lambda obj: added.append(obj)
    await _apply_many_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=UnitParent(id=1),
        relation_name="tags",
        rel_info=rel_with_link,
        relation_payload=[{"entity": {"id": 5}, "link": {"weight": 3}}],
    )
    assert any(isinstance(item, UnitLink) for item in added)

    # link update branch
    link_row = UnitLink(parent_id=1, tag_id=5, weight=1)

    async def _exec_row(_stmt: object) -> object:
        return SimpleNamespace(first=lambda: link_row)

    session.exec = _exec_row
    await _apply_many_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=UnitParent(id=1),
        relation_name="tags",
        rel_info=rel_with_link,
        relation_payload=[{"entity": {"id": 5}, "link": {"weight": 7}}],
    )
    assert link_row.weight == 7

    # link create policy denial
    deny_create = SimpleNamespace(
        can_create=lambda _u, _o: False,
        can_read=lambda _u, _o: True,
        can_update=lambda _u, _o, _i: True,
        can_delete=lambda _u, _o: True,
        filter_list_query=lambda _u, q: q,
    )
    ctx.nested_policy_registry[UnitLink] = cast(CrudPolicy, deny_create)
    session.exec = _exec_none
    with pytest.raises(ForbiddenError):
        await _apply_many_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="tags",
            rel_info=rel_with_link,
            relation_payload=[{"entity": {"id": 5}, "link": {"weight": 3}}],
        )

    # link update policy denial
    deny_update = SimpleNamespace(
        can_create=lambda _u, _o: True,
        can_read=lambda _u, _o: True,
        can_update=lambda _u, _o, _i: False,
        can_delete=lambda _u, _o: True,
        filter_list_query=lambda _u, q: q,
    )
    ctx.nested_policy_registry[UnitLink] = cast(CrudPolicy, deny_update)
    session.exec = _exec_row
    with pytest.raises(ForbiddenError):
        await _apply_many_to_many(
            ctx,
            parent_model=UnitParent,
            parent_obj=UnitParent(id=1),
            relation_name="tags",
            rel_info=rel_with_link,
            relation_payload=[{"entity": {"id": 5}, "link": {"weight": 7}}],
        )

    # link_model None without append branch (already present)
    parent_with_existing = SimpleNamespace(id=1, tags=[tag_obj])
    await _apply_many_to_many(
        ctx,
        parent_model=UnitParent,
        parent_obj=cast(SQLModel, parent_with_existing),
        relation_name="tags",
        rel_info=rel_without_link,
        relation_payload=[{"entity": {"id": 5}}],
    )
    assert len(parent_with_existing.tags) == 1


@pytest.mark.asyncio
async def test_apply_many_to_one_without_parent_object_does_not_set_attr(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    session = SimpleNamespace(get=AsyncMock(return_value=UnitChild(id=1)))
    ctx = _ctx(session)
    update_child = AsyncMock()
    monkeypatch.setattr("auen._nested_mutations._update_model_graph", update_child)

    await _apply_relation(
        ctx,
        parent_model=UnitParent,
        parent_obj=None,
        relation_name="author",
        rel_info=_rel(
            name="author",
            direction=RelationshipDirection.MANYTOONE,
            uselist=False,
            local_columns=("parent_fk",),
            target_model=UnitChild,
        ),
        relation_payload={"id": 1},
        parent_scalar_data={},
    )
    assert update_child.await_count == 1



"""Policy tests: verify row-level authorization."""

from __future__ import annotations

from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient


async def test_list_filtered_by_policy(policy_app: FastAPI) -> None:
    """LIST only returns rows owned by the current user."""
    transport = ASGITransport(app=policy_app)

    # Create books as alice
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "alice"},
    ) as alice:
        await alice.post(
            "/books/",
            json={
                "title": "Refactoring",
                "isbn": "978-0134757599",
                "owner_id": "alice",
            },
        )

    # Create books as bob
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "bob"},
    ) as bob:
        await bob.post(
            "/books/",
            json={
                "title": "Introduction to Algorithms",
                "isbn": "978-0262046305",
                "owner_id": "bob",
            },
        )

    # Alice should only see her books
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "alice"},
    ) as alice:
        resp = await alice.get("/books/")
        assert resp.status_code == 200
        books = resp.json()
        assert all(b["owner_id"] == "alice" for b in books)
        assert len(books) >= 1


async def test_read_forbidden_by_policy(policy_app: FastAPI) -> None:
    """READ on another user's object returns 403."""
    transport = ASGITransport(app=policy_app)

    # Create as bob
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "bob"},
    ) as bob:
        resp = await bob.post(
            "/books/",
            json={
                "title": "Structure and Interpretation of Computer Programs",
                "isbn": "978-0262510875",
                "owner_id": "bob",
            },
        )
        book_id = resp.json()["id"]

    # Alice tries to read bob's book
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "alice"},
    ) as alice:
        resp = await alice.get(f"/books/{book_id}")
        assert resp.status_code == 403


async def test_delete_forbidden_by_policy(policy_app: FastAPI) -> None:
    """DELETE on another user's object returns 403."""
    transport = ASGITransport(app=policy_app)

    # Create as bob
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "bob"},
    ) as bob:
        resp = await bob.post(
            "/books/",
            json={
                "title": "Compilers: Principles, Techniques, and Tools",
                "isbn": "978-0321486813",
                "owner_id": "bob",
            },
        )
        book_id = resp.json()["id"]

    # Alice tries to delete bob's book
    async with AsyncClient(
        transport=transport,
        base_url="http://test",
        headers={"X-User": "alice"},
    ) as alice:
        resp = await alice.delete(f"/books/{book_id}")
        assert resp.status_code == 403
